/*
 * Copyright (c) 2006 Atmel Corporation
 *
 * Based on the bttv driver for Bt848 with respective copyright holders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#define DEBUG
#include <linux/clk.h>
#include <linux/completion.h>
#include <linux/dma-mapping.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/videodev2.h>
#include <linux/wait.h>

// For isimem
#include <asm/arch/board.h>

#include <asm/io.h>

#include <media/v4l2-common.h>

#include "atmel-isi.h"

#define ATMEL_ISI_VERSION	KERNEL_VERSION(0, 1, 0)


// Marius added
//#define ISI_VIDEO_MAX_HSIZE 2048
//#define ISI_VIDEO_MAX_VSIZE 1536

#define ISI_VIDEO_MAX_HSIZE 1024
#define ISI_VIDEO_MAX_VSIZE 768


//#define ISI_VIDEO_MAX_MEM_SIZE 320*240*4
//#define ISI_VIDEO_MAX_MEM_SIZE 1024*768*2 // medium supported on ISI on greyscale 2 bytes / pixels
#define ISI_VIDEO_MAX_MEM_SIZE (ISI_VIDEO_MAX_HSIZE * ISI_VIDEO_MAX_VSIZE * 2) // max supported on ISI/image chip on greyscale 2 bytes / pixels = 6300k RAM

#define ISI_CAPTURE_MASK (ISI_BIT(SOF))

enum {
	STATE_IDLE = 0,
	STATE_CAPTURE_READY,
	STATE_CAPTURE_WAIT_SOF,
	STATE_CAPTURE_IN_PROGRESS,
	STATE_CAPTURE_DONE,
	STATE_CAPTURE_ERROR,
};

struct atmel_isi {
	void __iomem			*regs;

	int				state;
	void				*capture_buf;
	dma_addr_t			capture_phys;
	size_t				capture_buf_size;
	wait_queue_head_t		capture_wq;

	struct atmel_isi_camera		*camera;
	struct atmel_isi_format		format;

	struct mutex			mutex;
	int				users;

	struct video_device		vdev;
	struct completion		reset_complete;
	struct clk			*pclk;
	struct clk			*hclk;
	struct platform_device		*pdev;
	unsigned int			irq;
};
#define to_atmel_isi(vdev) container_of(vdev, struct atmel_isi, vdev)

struct atmel_isi_fh {
	struct atmel_isi		*isi;
	unsigned int			read_off;
};

/*
 * Interface to the actual camera.
 */
static LIST_HEAD(camera_list);
static DEFINE_MUTEX(camera_list_mutex);

static void avr32_isi_release_camera(struct atmel_isi *isi,
				     struct atmel_isi_camera *cam)
{
	/* FIXME: Terminate any ongoing transfers */
	mutex_lock(&camera_list_mutex);
	cam->isi = NULL;
	isi->camera = NULL;
	module_put(cam->owner);
	mutex_unlock(&camera_list_mutex);
}

int atmel_isi_register_camera(struct atmel_isi_camera *cam)
{
	pr_debug("atmel_isi: register camera %s\n", cam->name);

	mutex_lock(&camera_list_mutex);
	list_add_tail(&cam->list, &camera_list);
	mutex_unlock(&camera_list_mutex);

	return 0;
}
EXPORT_SYMBOL_GPL(atmel_isi_register_camera);

void atmel_isi_unregister_camera(struct atmel_isi_camera *cam)
{
	pr_debug("atmel_isi: unregister camera %s\n", cam->name);

	mutex_lock(&camera_list_mutex);
	if (cam->isi)
		cam->isi->camera = NULL;
	list_del(&cam->list);
	mutex_unlock(&camera_list_mutex);
}
EXPORT_SYMBOL_GPL(atmel_isi_unregister_camera);

static struct atmel_isi_camera *
avr32_isi_grab_camera(struct atmel_isi *isi)
{
	struct atmel_isi_camera *entry, *cam = NULL;

	mutex_lock(&camera_list_mutex);
	list_for_each_entry(entry, &camera_list, list) {
		/* Just grab the first camera available */
		if (!entry->isi) {
			if (!try_module_get(entry->owner))
				continue;

			cam = entry;
			cam->isi = isi;
			pr_debug("%s: got camera: %s\n",
				 isi->vdev.name, cam->name);
			break;
		}
	}
	mutex_unlock(&camera_list_mutex);

	return cam;
}

static int avr32_isi_set_format(struct atmel_isi *isi,
				struct atmel_isi_format *fmt)
{
	struct atmel_isi_camera *cam = isi->camera;
	unsigned int isi_hsize;
	u32 cr2;
	int ret;

	switch (fmt->bytes_per_pixel) {
	case 1:
	case 2:
		cr2 = ISI_BF(RGB_MODE, 1);
		break;
	case 3:
		cr2 = ISI_BF(RGB_MODE, 0);
		break;
	default:
		return -EINVAL;
	}
		
	fmt->pix.width = min(ISI_VIDEO_MAX_HSIZE, fmt->pix.width);
	fmt->pix.height = min(ISI_VIDEO_MAX_VSIZE, fmt->pix.height);
	fmt->pix.bytesperline = 0;

	ret = cam->set_format(cam, fmt);
	if (ret)
		return ret;

	fmt->pix.bytesperline
		= ALIGN(fmt->pix.width * fmt->bytes_per_pixel, 4);
	fmt->pix.sizeimage = fmt->pix.bytesperline * fmt->pix.height;

	pr_debug("pix.bytesperline: %i pix.sizeimage: %i \n", fmt->pix.bytesperline, fmt->pix.sizeimage);


	isi_hsize = fmt->pix.width;
	/* if (fmt->bytes_per_pixel == 1)
	   isi_hsize /= 2; */

	cr2 |= (ISI_BF(IM_VSIZE, fmt->pix.height - 1)
		| ISI_BF(IM_HSIZE, isi_hsize - 1)
		/* FIXME */
		| ISI_BF(RGB_SWAP, 0)
		| ISI_BF(COL_SPACE, 0)
		| ISI_BF(RGB_CFG, 0)

		// Marius change. Change to compact grayscale mode. Should be found from requested format from v4l2
		| ISI_BF(GRAYSCALE, 1)
		| ISI_BF(GS_MODE, 0) );
		

	isi_writel(isi, CR1, (ISI_BIT(DIS)
			      /* | ISI_BF(SLD, 202) */
			      /* | ISI_BF(SFD, 16) */
			      | ISI_BF(EMB_SYNC, 0)
			      | ISI_BF(HSYNC_POL, cam->hsync_act_low)
			      | ISI_BF(VSYNC_POL, cam->vsync_act_low)
			      | ISI_BF(PIXCLK_POL, cam->pclk_act_falling)));
	isi_writel(isi, CR2, cr2);

	return 0;
}

/* Get the default format from the camera and modify it to suit our needs. */
static int avr32_isi_init_format(struct atmel_isi *isi)
{
  //  pr_debug( "start of avr32_isi_init_format\n" );
	struct atmel_isi_camera *cam = isi->camera;
	struct atmel_isi_format fmt;
	int ret;

	ret = cam->get_format(cam, &fmt);
	if (ret)
		return ret;

	pr_debug( "got formats in avr32_isi_init_format\n" );

	ret = avr32_isi_set_format(isi, &fmt);

	pr_debug( "set formats in avr32_isi_init_format res: %i\n", ret );

	if (!ret)
		memcpy(&isi->format, &fmt, sizeof(struct atmel_isi_format));

	return ret;
}

static int avr32_isi_start_capture(struct atmel_isi *isi)
{
	u32 cr1;
	int ret;

	/* Wait for SOF */
	pr_debug("isi: waiting for SOF\n");
	isi->state = STATE_IDLE;
	cr1 = isi_readl(isi, CR1);
	cr1 &= ~ISI_BIT(DIS);
	isi_writel(isi, CR1, cr1);
	isi_writel(isi, IER, ISI_BIT(SOF));
	ret = wait_event_interruptible(isi->capture_wq,
				       isi->state != STATE_IDLE);
	if (ret)
		return ret;
	if (isi->state != STATE_CAPTURE_READY)
		return -EIO;

	/*
	 * Do a codec request. Next SOF indicates start of capture,
	 * the one after that indicates end of capture.
	 */
	pr_debug("isi: starting capture\n");
	isi_writel(isi, CDBA, isi->capture_phys);

	isi->state = STATE_CAPTURE_WAIT_SOF;
	cr1 = isi_readl(isi, CR1);

	// Marius test without capture	cr1 |= ISI_BIT(CODEC_ON);
	cr1 |= ISI_BIT(CODEC_ON);

	isi_writel(isi, CR1, cr1);

	isi_writel(isi, IER, ISI_CAPTURE_MASK);

	return 0;
}

#if 0
static int avr32_isi_init_hw(struct atmel_isi *isi, struct avr32_4vl2_dev *dev, struct avr32_v4l2_ops *ops)
{
	unsigned int cr1_reg_val, cr2_reg_val;

	/* Reset ISI module */
	isi_writel(isi, CR1, ISI_BIT(RST));

	/*Read and discard reset interrupt*/
	isi_readl(isi,SR);

	/* Setup interface settings */
	cr1_reg_val = \
		ISI_BF(HSYNC_POL, dev->hsync_pol) | \
		ISI_BF(VSYNC_POL, dev->vsync_pol) | \
		ISI_BF(PIXCLK_POL, dev->pixclk_pol) | \
		ISI_BF(EMB_SYNC, dev->emb_sync) | \
		ISI_BF(CRC_SYNC, dev->crc_sync) | \
		ISI_BF(FRATE, ops->frate) | \
		ISI_BF(FULL, dev->full) | \
		ISI_BF(THMASK, dev->thmask) | \
		ISI_BIT(DIS) | \
		ISI_BF(SLD, dev->sld) | \
		ISI_BF(SFD, dev->sfd);

	cr2_reg_val = \
		ISI_BF(IM_VSIZE, ops->im_vsize-1) | \
		ISI_BF(GS_MODE, ops->gs_mode) | \
		ISI_BF(RGB_MODE, ops->rgb_mode) | \
		ISI_BF(GRAYSCALE, ops->gs) | \
		ISI_BF(RGB_SWAP, ops->rgb_swap) | \
		ISI_BF(COL_SPACE, ops->col_space) | \
		ISI_BF(IM_HSIZE, ops->im_hsize-1) | \
		ISI_BF(YCC_SWAP, ops->ycc_swap) | \
		ISI_BF(RGB_CFG, ops->rgb_cfg );

	isi_writel(isi, CR1, cr1_reg_val);
	isi_writel(isi, CR2, cr2_reg_val);

	return 0;
}
#endif

/* Called when all references to vdev have been dropped */
static void avr32_isi_release (struct video_device *vdev)
{
	struct atmel_isi *isi = to_atmel_isi(vdev);
	pr_debug("%s: release\n", vdev->name);
	kfree(isi);
}

#if 0
int avr32_isi_reqbufs(struct atmel_isi_fh *isi_fh, struct v4l2_requestbuffers *req)
{
	int retval=0, i;

	if (( req->count * isi_fh->fmt.fmt.pix.width * isi_fh->fmt.fmt.pix.height) > ISI_VIDEO_MAX_MEM_SIZE ){
		req->count = ISI_VIDEO_MAX_MEM_SIZE / \
			(isi_fh->fmt.fmt.pix.width * isi_fh->fmt.fmt.pix.height);
		printk(KERN_INFO "The number of bufs requested exceeds the \
			maximum memory allocated within the driver. Thus %d \
			buffer%s is the maximum with the given resolution. \
			%d buffer%s is given\n", req->count, req->count > 1 ? "s" : "", \
			req->count, req->count > 1 ? "s" : "");
	}
	isi_fh->count=req->count;
	isi_fh->bufsize = isi_fh->fmt.fmt.pix.width * isi_fh->fmt.fmt.pix.height;
	isi_fh->nextbuf=-1;

	for(i=0; i<isi_fh->count; i++)
		isi_fh->active_bufs[i] = 0;
	isi_fh->mapped_bufs=0;

	return retval;
}


static int isi_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct atmel_isi_fh *isi_fh = (struct atmel_isi_fh *) file->private_data;

	vma->vm_flags |= VM_RESERVED;
	vma->vm_flags = VM_DONTEXPAND; /* fixed size */
	vma->vm_flags &= ~VM_IO;  /* shared anonumous pages */

	if(isi_fh->mapped_bufs){
		if (remap_pfn_range(vma,vma->vm_start,(isi_fh->phys_addr>>PAGE_SHIFT),ISI_VIDEO_MAX_MEM_SIZE,vma->vm_page_prot))
			return -EAGAIN;
	}

	// printk(KERN_INFO "vm start: 0x%08lx, vm range: 0x%08lx\n", vma->vm_start, vma->vm_end-vma->vm_start );

	isi_fh->mapped_bufs++;

	return 0;
}
#endif

static void avr32_isi_capture_done(struct atmel_isi *isi,
				   int state)
{
	u32 cr1;

	cr1 = isi_readl(isi, CR1);
	cr1 &= ~ISI_BIT(CODEC_ON);
	cr1 |= ISI_BIT(DIS);
	isi_writel(isi, CR1, cr1);

	isi->state = state;
	wake_up_interruptible(&isi->capture_wq);
	isi_writel(isi, IDR, ISI_CAPTURE_MASK);

	pr_debug("isi: finished capture\n");
}

/* isi interrupt function */
static irqreturn_t isi_interrupt(int irq, void *dev_id)
{
	struct atmel_isi *isi = dev_id;
	u32 status, mask, pending;
	irqreturn_t ret = IRQ_NONE;

	status = isi_readl(isi, SR);
	mask = isi_readl(isi, IMR);
	pending = status & mask;

	while (pending) {
		if (pending & (ISI_BIT(FO_C_OVF) | ISI_BIT(FR_OVR))) {
			avr32_isi_capture_done(isi, STATE_CAPTURE_ERROR);
			pr_debug("%s: FIFO overrun (status=0x%x)\n",
				 isi->vdev.name, status);
		} else if (pending & ISI_BIT(SOF)) {
			switch (isi->state) {
			case STATE_IDLE:
				isi->state = STATE_CAPTURE_READY;
				wake_up_interruptible(&isi->capture_wq);
				break;
			case STATE_CAPTURE_READY:
				break;
			case STATE_CAPTURE_WAIT_SOF:
				isi->state = STATE_CAPTURE_IN_PROGRESS;
				break;
			case STATE_CAPTURE_IN_PROGRESS:
				avr32_isi_capture_done(isi, STATE_CAPTURE_DONE);
				break;
			}
		}
		if (pending & ISI_BIT(SOFTRST)) {
			complete(&isi->reset_complete);
			isi_writel(isi, IDR, ISI_BIT(SOFTRST));
		}

		status = isi_readl(isi, SR);
		mask = isi_readl(isi, IMR);
		pending = status & mask;
		ret = IRQ_HANDLED;
	}

	return ret;
}

#if 0
/* This function waits for a complete frame to written to memory by the isi module */
int avr32_isi_buf_int_capture(struct atmel_isi *isi, unsigned int phys_addr)
{
	int val, timeout=msecs_to_jiffies(CAM_TIMEOUT);
	int retval = 1;

	init_completion(&interrupt_completion);

	isi_writel(isi, CDBA, phys_addr);
	isi_writel(isi, IER, ~0UL);
	val = isi_readl(isi, CR1);
	val |= ISI_BIT(CODEC_ON); /*enable codec path */
	val &= ~ISI_BIT(DIS); /* enable interface */
	isi_writel(isi, CR1, val);

	if(!wait_for_completion_timeout(&interrupt_completion, timeout))
		retval = -EIO;

	capturing = 0;
	val  = isi_readl(isi, CR1);
	val &= ~ISI_BIT(CODEC_ON); /* disable codec path */
	val |= ISI_BIT(DIS); /*finish capturing current frame, then shutdown */
	isi_writel(isi, CR1, val);

	return retval;

}

static int isi_getnextbuf(struct atmel_isi_fh *isi_fh)
{

	if ( (isi_fh->nextbuf < 0) | ((isi_fh->nextbuf+1)>=isi_fh->count) )
		isi_fh->nextbuf=0;
	else
		isi_fh->nextbuf++;
//	printk(KERN_INFO "atmel-isi: The next buffer to use is #%d of #%d available\n", isi_fh->nextbuf, isi_fh->count);

	return 0;
}

static int avr32_isi_getbuf(struct atmel_isi_fh *isi_fh)
{
		isi_fh->active_bufs[isi_fh->nextbuf] = 0;

	return 0;
}


static int avr32_isi_querybuf(struct atmel_isi_fh *isi_fh, struct v4l2_buffer *b)
{
	b->flags = 0;
	b->length = isi_fh->fmt.fmt.pix.width * isi_fh->fmt.fmt.pix.height;
	b->bytesused = b->length;
	b->m.offset = b->index*b->length;
	b->input = isi_fh->virtual_addr;

	return 0;
}

/* Used to get a memory mapped isi buffer, called with IO_METHOD_MMAP */
static int avr32_isi_dqbuf(struct atmel_isi_fh *isi_fh, struct v4l2_buffer *b)
{
	/* fixme */
	flush_cache_all();

	isi_getnextbuf(isi_fh);
	isi_fh->active_bufs[isi_fh->nextbuf] = 1;
/*
	printk(KERN_INFO "atmel-isi: physical base address @ 0x%08x\n", isi_fh->phys_addr);
	printk(KERN_INFO "atmel-isi: virtual base address @ 0x%08x\n", isi_fh->virtual_addr);
	printk(KERN_INFO "atmel-isi: next buffer: %d\n", isi_fh->nextbuf);
	printk(KERN_INFO "atmel-isi: buffer size: 0x%06x\n", isi_fh->bufsize);
*/
//	printk(KERN_INFO "atmel-isi: Using buffer #%d @ 0x%08x with length 0x%08x\n", isi_fh->nextbuf, isi_fh->phys_addr+(isi_fh->nextbuf * isi_fh->bufsize), isi_fh->bufsize );

	b->length = isi_fh->bufsize;
	b->index = isi_fh->nextbuf;
	b->m.offset = isi_fh->virtual_addr+(isi_fh->nextbuf * isi_fh->bufsize);

	return avr32_isi_buf_int_capture(isi_fh->isi,
					 isi_fh->phys_addr+(isi_fh->nextbuf * b->length) );

}
#endif

static int avr32_isi_querycap(struct file *file, void *priv,
			      struct v4l2_capability *cap)
{
	strcpy(cap->driver, "atmel-isi");
	strcpy(cap->card, "Atmel Image Sensor Interface");
	cap->version = ATMEL_ISI_VERSION;
	cap->capabilities = (V4L2_CAP_VIDEO_CAPTURE
			     | V4L2_CAP_READWRITE);

	return 0;
}

static int avr32_isi_enum_input(struct file *file, void *priv,
				struct v4l2_input *input)
{
	struct atmel_isi_fh *fh = priv;
	struct atmel_isi *isi = fh->isi;

	if (input->index != 0)
		return -EINVAL;

	strlcpy(input->name, isi->camera->name, sizeof(input->name));
	input->type = V4L2_INPUT_TYPE_CAMERA;
	input->std = V4L2_STD_PAL;

	return 0;
}

static int avr32_isi_s_input(struct file *file, void *priv,
			     unsigned int index)
{
	if (index != 0)
		return -EINVAL;
	return 0;
}

static int avr32_isi_g_input(struct file *file, void *priv,
			     unsigned int *index)
{
	*index = 0;
	return 0;
}

static int avr32_isi_enum_fmt_cap(struct file *file, void *priv,
				  struct v4l2_fmtdesc *fmt)
{
	struct atmel_isi_fh *fh = priv;
	struct atmel_isi *isi = fh->isi;

	return isi->camera->enum_format(isi->camera, fmt);
}

static int avr32_isi_g_fmt_cap(struct file *file, void *priv,
			       struct v4l2_format *vfmt)
{
	struct atmel_isi_fh *fh = priv;
	struct atmel_isi *isi = fh->isi;

	memcpy(&vfmt->fmt.pix, &isi->format.pix,
	       sizeof(struct v4l2_pix_format));

	return 0;
}

static int avr32_isi_s_fmt_cap(struct file *file, void *priv,
			       struct v4l2_format *vfmt)
{
	struct atmel_isi_fh *fh = priv;
	struct atmel_isi *isi = fh->isi;
	int ret;

	memcpy(&isi->format.pix, &vfmt->fmt.pix,
	       sizeof(struct v4l2_pix_format));
	ret = avr32_isi_set_format(isi, &isi->format);
	if (!ret)
		memcpy(&vfmt->fmt.pix, &isi->format.pix,
		       sizeof(struct v4l2_pix_format));

	return ret;
}


/** TODO: Marius: old? **/

#if 0
static int avr32_isi_request_features(struct atmel_isi_fh *isi_fh, struct v4l2_format *req_fmt)
{
  // TODO: this should consult the camera instead of using cam_ops in atmel-isi.h...

	if ( (req_fmt->fmt.pix.width <= cam_ops.im_hsize) && \
	(req_fmt->fmt.pix.height <= cam_ops.im_vsize) ) {
		isi_fh->fmt.fmt.pix.width = req_fmt->fmt.pix.width;
		isi_fh->fmt.fmt.pix.height = req_fmt->fmt.pix.height;
	}
	else {
	  	pr_debug("avr32_isi_request_features returns invalid image size...");
		return -ENOMEM;
	}

	if (req_fmt->fmt.pix.pixelformat & cam_ops.format)
		isi_fh->fmt.fmt.pix.pixelformat = req_fmt->fmt.pix.pixelformat;
	else {
	  pr_debug("avr32_isi_request_features returns unsupported format...");
	  return -EIO;
	}

	pr_debug("avr32_isi_request_features succeded");


	/*
	 * other feature checks may be done here, but the features needs to be
	 * implemented...
	 */
/*
	isi_fh->fmt.timecode.type = V4L2_TC_TYPE_24FPS;
	isi_fh->fmt.timecode.frames = 1;
*/
	return 0;
}

static int avr32_isi_do_ioctl (struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
	struct atmel_isi_fh *isi_fh = (struct atmel_isi_fh *) file->private_data;

	switch(cmd){
		case VIDIOC_S_FMT:{
		  	pr_debug("VIDIOC_S_FMT called");

			struct v4l2_format *fmt = arg;
			return avr32_isi_request_features(isi_fh, fmt);
			}
		case VIDIOC_REQBUFS:
			return avr32_isi_reqbufs(isi_fh, arg);

		case VIDIOC_QUERYBUF:
			return avr32_isi_querybuf(isi_fh,arg);

		case VIDIOC_QBUF:
			return avr32_isi_getbuf(isi_fh);

		case VIDIOC_DQBUF:
			return avr32_isi_dqbuf(isi_fh,arg);

		default:
		  pr_debug("Warning: unsupported v4l2 command %i called\n", cmd);
			return -EIO;
	}
}
#endif

static int avr32_isi_close (struct inode *inode, struct file *file)
{
	struct atmel_isi_fh *fh = file->private_data;
	struct atmel_isi *isi = fh->isi;

	mutex_lock(&isi->mutex);
	isi->users--;
	kfree(fh);

	/* TODO: Stop any ongoing capture */

	mutex_unlock(&isi->mutex);

	return 0;
}

static int avr32_isi_open (struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct atmel_isi *isi = to_atmel_isi(vdev);
	struct atmel_isi_fh *fh;
	int ret = -EBUSY;

	mutex_lock(&isi->mutex);

	if (isi->users) {
		pr_debug("%s: open(): device busy\n", vdev->name);
		goto out;
	}

	if (!isi->camera) {
		unsigned long timeout;

		ret = -ENODEV;
		isi->camera = avr32_isi_grab_camera(isi);
		if (!isi->camera)
			goto out;
		/*
		 * Reset the controller and wait for completion. The
		 * reset will only succeed if we have a pixel clock
		 * from the camera.
		 */
		init_completion(&isi->reset_complete);
		isi_writel(isi, CR1, ISI_BIT(RST) | ISI_BIT(DIS));
		isi_writel(isi, IER, ISI_BIT(SOFTRST));
		timeout = wait_for_completion_timeout(&isi->reset_complete,
						      msecs_to_jiffies(5000)); // Marius change: longer timeout (was 100)

		if (timeout) {
			// Marius test: skip this temporary: 
		  //ret = -ETIMEDOUT;
		  // isi->camera = 0; // Marius change: deselect camera
			// goto out;
		  pr_debug( "Timout but ignoring for testing..." );
		}
		isi_writel(isi, IDR, ~0UL);


		pr_debug( "Finished resetting ISI. Initializing format\n" );
		avr32_isi_init_format(isi);

		pr_debug( "Starting capture (in isi_open)\n" );

		isi->camera->start_capture(isi->camera);
	}

	ret = -ENOMEM;
	fh = kzalloc(sizeof(struct atmel_isi_fh), GFP_KERNEL);
	if (!fh) {
		pr_debug("%s: open(): out of memory\n", vdev->name);
		goto out;
	}

	fh->isi = isi;
	file->private_data = fh;
	isi->users++;

	ret = 0;

out:
	mutex_unlock(&isi->mutex);

	return ret;
}

static ssize_t avr32_isi_read(struct file *file, char __user *data,
			      size_t count, loff_t *ppos)
{
	struct atmel_isi_fh *fh = file->private_data;
	struct atmel_isi *isi = fh->isi;
	int state;
	int ret;

	state = isi->state;
	barrier();

	pr_debug("isi_read called\n");

	if (state != STATE_CAPTURE_DONE) {
	  pr_debug("inside isi_read: starting capture\n" );
		if (state == STATE_IDLE)
			avr32_isi_start_capture(isi);

		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;

		ret = wait_event_interruptible(
			isi->capture_wq,
			(isi->state == STATE_CAPTURE_DONE)
			|| (isi->state == STATE_CAPTURE_ERROR));
		if (ret)
			return ret;
		if (isi->state == STATE_CAPTURE_ERROR) {
			isi->state = STATE_IDLE;
			return -EIO;
		}

		fh->read_off = 0;
	}

	count = min(count, (size_t)isi->format.pix.sizeimage - fh->read_off);
	ret = copy_to_user(data, isi->capture_buf + fh->read_off, count);

	pr_debug("inside isi_read: copy count: %i, ret: %i\n", count, ret);

	if (ret)
		return -EFAULT;

	fh->read_off += count;
	if (fh->read_off >= isi->format.pix.sizeimage)
		isi->state = STATE_IDLE;

	return count;
}

static struct file_operations avr32_isi_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= video_ioctl2,
	.open		= avr32_isi_open,
	.release	= avr32_isi_close,
	.read		= avr32_isi_read,
	.compat_ioctl	= v4l_compat_ioctl32,
};

static int __devexit avr32_isi_remove(struct platform_device *pdev)
{
	struct atmel_isi *isi = platform_get_drvdata(pdev);

	if (isi->camera)
		isi->camera->stop_capture(isi->camera);

	// Marius fix for preallocated buffer
	struct atmel_isi_platform_data *isi_mem_data = pdev->dev.platform_data;

	// Only free memory if allocated. If preallallocated, just leave it alone
        if (!isi_mem_data || isi_mem_data->isimem_size == 0) {
	  dma_free_coherent(&pdev->dev, isi->capture_buf_size,
			  isi->capture_buf, isi->capture_phys);
	}
	else { // unmap preallocated RAM
	  iounmap( isi->capture_buf );
	}

	platform_set_drvdata(pdev, NULL);

        free_irq(isi->irq, isi);
	iounmap(isi->regs);
	clk_disable(isi->hclk);
	clk_disable(isi->pclk);
	clk_put(isi->hclk);
	clk_put(isi->pclk);

	if (isi->camera)
		avr32_isi_release_camera(isi, isi->camera);
	video_unregister_device(&isi->vdev);

	/*
	 * Don't free isi here -- it will be taken care of by the
	 * release() callback.
	 */

	return 0;
}


static int __devinit avr32_isi_probe(struct platform_device *pdev)
{
	unsigned int irq;
	struct atmel_isi *isi;
	struct clk *pclk, *hclk;
	struct resource *regs;
	int ret;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(!regs)
		return -ENXIO;

        pclk = clk_get(&pdev->dev, "pclk");
        if (IS_ERR(pclk))
		return PTR_ERR(pclk);
        hclk = clk_get(&pdev->dev, "hclk");
        if (IS_ERR(hclk)) {
		ret = PTR_ERR(hclk);
		goto err_hclk;
	}
	clk_enable(pclk);
	clk_enable(hclk);

	isi = kzalloc(sizeof(struct atmel_isi), GFP_KERNEL);
	if(!isi){
		ret = -ENOMEM;
		dev_err(&pdev->dev, "can't allocate interface!\n");
		goto err_alloc_isi;
	}

	isi->pclk = pclk;
	isi->hclk = hclk;
	isi->capture_buf_size = ISI_VIDEO_MAX_MEM_SIZE;
	mutex_init(&isi->mutex);
	init_waitqueue_head(&isi->capture_wq);

	isi->vdev.fops = &avr32_isi_fops;
	strcpy(isi->vdev.name, "atmel-isi");
	isi->vdev.type = VFL_TYPE_GRABBER,
	isi->vdev.type2 = VID_TYPE_CAPTURE;
	isi->vdev.minor = -1;
	isi->vdev.release = avr32_isi_release;
#ifdef DEBUG
	isi->vdev.debug = V4L2_DEBUG_IOCTL | V4L2_DEBUG_IOCTL_ARG;
#endif

	/* install ioctl callbacks */
	isi->vdev.vidioc_querycap = avr32_isi_querycap;
	isi->vdev.vidioc_enum_fmt_cap = avr32_isi_enum_fmt_cap;
	isi->vdev.vidioc_g_fmt_cap = avr32_isi_g_fmt_cap;
	isi->vdev.vidioc_s_fmt_cap = avr32_isi_s_fmt_cap;
	isi->vdev.vidioc_enum_input = avr32_isi_enum_input;
	isi->vdev.vidioc_g_input = avr32_isi_g_input;
	isi->vdev.vidioc_s_input = avr32_isi_s_input;

	isi->regs = ioremap(regs->start, regs->end - regs->start + 1);
        if (!isi->regs) {
		ret = -ENOMEM;
		goto err_ioremap;
	}

	irq = platform_get_irq(pdev,0);
	ret = request_irq(irq, isi_interrupt, 0, "isi", isi);
	if (ret) {
		dev_err(&pdev->dev, "unable to request irq %d\n", irq);
		goto err_req_irq;
	}
	isi->irq = irq;

	// Marius test, use framebuffer memory:

	struct atmel_isi_platform_data *isi_mem_data = pdev->dev.platform_data;

        /* Use platform-supplied framebuffer memory if available */
        if (isi_mem_data && isi_mem_data->isimem_size != 0) {
	  printk( "Using preallocated isimem.\n" );
	  
	  isi->capture_phys = isi_mem_data->isimem_start;
	  isi->capture_buf_size = isi_mem_data->isimem_size;

	  isi->capture_buf = ioremap( isi->capture_phys, isi->capture_buf_size );
        } else {
		printk( "No preallocated isimem found, trying to allocate ISI buffer now\n" );

		/* Allocate capture buffer */
		isi->capture_buf = dma_alloc_coherent(&pdev->dev,
					      isi->capture_buf_size,
					      &isi->capture_phys,
				      GFP_KERNEL);

        }

	if (!isi->capture_buf) {
		ret = -ENOMEM;
		dev_err(&pdev->dev, "failed to allocate capture buffer\n");
		goto err_alloc_buf;
	}

	//	memset(isi->capture_buf, 0, isi->capture_buf_size);
	// Marius test
	memset(isi->capture_buf, 65, isi->capture_buf_size); // set in a

	dev_info(&pdev->dev,
		 "capture buffer: %zu bytes at %p (phys 0x%08x)\n",
		 isi->capture_buf_size, isi->capture_buf,
		 isi->capture_phys);

	isi_writel(isi, CR1, ISI_BIT(DIS));

        /* Register Video Device */
	ret = video_register_device(&isi->vdev, VFL_TYPE_GRABBER, -1);
	if (ret)
                goto err_register;

	platform_set_drvdata(pdev, isi);

	dev_info(&pdev->dev, "Atmel ISI V4L2 device at 0x%08lx\n",
		 (unsigned long)regs->start);

	return 0;

err_register:
	dma_free_coherent(&pdev->dev, isi->capture_buf_size,
			  isi->capture_buf, isi->capture_phys);
err_alloc_buf:
        free_irq(isi->irq, isi);
err_req_irq:
	iounmap(isi->regs);
err_ioremap:
	kfree(isi);
err_alloc_isi:
	clk_disable(hclk);
	clk_disable(pclk);
	clk_put(hclk);
err_hclk:
	clk_put(pclk);
	return ret;

}

static struct platform_driver avr32_isi_driver = {
	.probe		= avr32_isi_probe,
	.remove		= __devexit_p(avr32_isi_remove),
	.driver		= {
		.name = "atmel_isi",
		.owner = THIS_MODULE,
	},
};

static int __init avr32_isi_init(void)
{
	return  platform_driver_register(&avr32_isi_driver);
}


static void __exit avr32_isi_exit(void)
{
	platform_driver_unregister(&avr32_isi_driver);
}


module_init(avr32_isi_init);
module_exit(avr32_isi_exit);

MODULE_AUTHOR("Espen Krangnes <ekrangnes@atmel.com>");
MODULE_DESCRIPTION("The V4L2 driver for AVR32 Linux");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("video");
