#include "stepper.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include <math.h>

#define DEBUG 1

//typedef short boolean;
typedef enum {false = 0, true = 1} boolean;

typedef struct stepper_full_status1_t {
  unsigned int address         : 8;

  unsigned int i_run           : 4;
  unsigned int i_hold          : 4;

  unsigned int v_max           : 4;
  unsigned int v_min           : 4;

  unsigned int acc_shape       : 1;
  unsigned int step_mode       : 2;
  unsigned int shaft           : 1;
  unsigned int acceleration    : 4;

  unsigned int vdd_reset       : 1;
  unsigned int step_loss       : 1;
  unsigned int electrical_def  : 1;
  unsigned int uv2             : 1;
  unsigned int thermal_shutd   : 1;
  unsigned int thermal_warn    : 1;
  unsigned int temp_info       : 2;

  unsigned int motion_status   : 3;
  unsigned int ext_sw_status   : 1;
  unsigned int over_current_1  : 1;
  unsigned int over_current_2  : 1;
  unsigned int stall_detected  : 1;
  unsigned int charge_pump_err : 1;

  unsigned int                 : 8;
  unsigned int abs_threshold   : 4;
  unsigned int delta_threshold : 4;
} stepper_full_status1_t;

/**
Buf min 8 bytes
**/
short stepper_read_status1( stepper_full_status1_t* status1 )
{
	int file;
	int reg = 0x81; // GetFullStatus1
	unsigned char wbuf[1] = { (unsigned char)reg };
	struct i2c_msg t_testSet[2] = {{STEPPER_ADDRESS, 0, 1, wbuf},{STEPPER_ADDRESS, 1, 8, (unsigned char*)status1}};
	struct i2c_rdwr_ioctl_data t_msgSet;
	t_msgSet.msgs = t_testSet;
	t_msgSet.nmsgs = 2;

        if ((file = open(TWI_DEVICE,O_RDWR)) < 0) return -ENODEV;

	if( ioctl(file,I2C_RDWR,&t_msgSet) < 0 ) return -1;

	return close(file);
}

typedef struct stepper_full_status2_t {
  unsigned int address         : 8;
  signed   int act_pos         : 16;
  signed   int target_pos      : 16;

  unsigned int secure_pos_low  : 8;

  unsigned int FS2StallEn      : 3;
  unsigned int                 : 1;
  unsigned int DC100           : 1;
  unsigned int secure_pos_high : 3;

  unsigned int abs_stall        : 1;
  unsigned int delta_stall_low  : 1;
  unsigned int delta_stall_high : 1;
  unsigned int min_samples      : 3;
  unsigned int dc100_stall_en   : 1;
  unsigned int pwmj_en          : 1;
} stepper_full_status2_t;

short stepper_read_status2( stepper_full_status2_t* status )
{
	int file;
	int reg = 0xFC; // GetFullStatus2
	unsigned char wbuf[1] = { (unsigned char)reg };
	struct i2c_msg t_testSet[2] = {{STEPPER_ADDRESS, 0, 1, wbuf},{STEPPER_ADDRESS, 1, 8, (unsigned char*)status}};
	struct i2c_rdwr_ioctl_data t_msgSet;
	t_msgSet.msgs = t_testSet;
	t_msgSet.nmsgs = 2;

        if ((file = open(TWI_DEVICE,O_RDWR)) < 0) return -ENODEV;

	if( ioctl(file,I2C_RDWR,&t_msgSet) < 0 ) return -1;

	return close(file);
}

int stepper_set_position(int pos)
{
        int file;
        unsigned char msg_data[5]= {0};
        struct i2c_msg msg = {STEPPER_ADDRESS, 0, 5, msg_data};
        struct i2c_rdwr_ioctl_data wr = { &msg, 1 };
	short reg = 0x8B; // SetPosition

	pos <<= 3; // half stepping

        /* open handle */
        if ((file = open(TWI_DEVICE,O_RDWR)) < 0)
                return -ENODEV;
	
	msg_data[0] = (unsigned char) reg;
	
	msg_data[1] = 0xFF;
	msg_data[2] = 0xFF;
	msg_data[3] = (unsigned char) ((pos&0xFF00)>>8);
	if ( pos < 0 ) msg_data[3] |= 0x80; // set sign bit
	msg_data[4] = (unsigned char) (pos&0x00FF);

        if ( ioctl(file, I2C_RDWR, &wr) <0 )
                return -1;
	close(file);

	return 0;
}

int stepper_set_motor_params(unsigned char iRun, unsigned char iHold)
{
        int file;
        unsigned char msg_data[8]= {0};
        struct i2c_msg msg = {STEPPER_ADDRESS, 0, 8, msg_data};
        struct i2c_rdwr_ioctl_data wr = { &msg, 1 };
        short reg = 0x89; // SetMotorParams
        /* open handle */
        if ((file = open(TWI_DEVICE,O_RDWR)) < 0)
                return -ENODEV;
	

	msg_data[0] = (unsigned char) reg;
	msg_data[1] = 0xFF;
	msg_data[2] = 0xFF;
	msg_data[3] = (unsigned char) (((iRun&0x0F) << 4) | ((iHold & 0x0F)));
	msg_data[4] = 0x01; // Vmin/Vmax to min.
	msg_data[5] = 0x00; // Sec position, acc to min.
	msg_data[6] = 0x00; // Sec pos.
	msg_data[7] = 0xA2; // Div. set to default

        if ( ioctl(file, I2C_RDWR, &wr) <0 )
                return -1;
	close(file);

	return 0;
}

typedef struct set_stall_param_t {
  unsigned int command         : 8; // = 0x96
  unsigned int data1           : 8;
  unsigned int data2           : 8;

  unsigned int Irun            : 4;
  unsigned int Ihold           : 4;

  unsigned int Vmax            : 4;
  unsigned int Vmin            : 4;

  unsigned int min_samples     : 3;
  unsigned int shaft           : 1;
  unsigned int acc             : 4;

  unsigned int abs_threshold   : 4;
  unsigned int delta_threshold : 4;

  unsigned int fs2stall_en     : 3;
  unsigned int acc_shape       : 1;
  unsigned int step_mode       : 2;
  unsigned int dc100_st_en     : 1;
  unsigned int pwmj_en         : 1;
} set_stall_param_t;

int stepper_set_stall_params(unsigned char iRun, unsigned char iHold)
{
        int file;
        set_stall_param_t msg_data = {0};
        struct i2c_msg msg = {STEPPER_ADDRESS, 0, 8, (unsigned char*)&msg_data};
        struct i2c_rdwr_ioctl_data wr = { &msg, 1 };
        short reg = 0x96; // SetStallParam
        /* open handle */
        if ((file = open(TWI_DEVICE,O_RDWR)) < 0)
                return -ENODEV;

	msg_data.command = (unsigned char) reg;
	msg_data.data1 = 0xFF;
	msg_data.data2 = 0xFF;
	
	msg_data.Irun = iRun;
	msg_data.Ihold = iHold;

	msg_data.Vmax = 0; // Vmax to min.
	msg_data.Vmin = 1; // Vmin to min. part of Vmax

	msg_data.min_samples = 0; // (?)
	msg_data.shaft = 1; // shaft direction
	msg_data.acc = 0x00; // acceleration to min.
	
	msg_data.abs_threshold = 0x2; // Enable and try highest (sensitivity)
	msg_data.delta_threshold = 0x00; // Disable...  // Enable and try highest threshold
	
	msg_data.fs2stall_en = 0; // trying lowest
	msg_data.acc_shape = 0; // normal accelecation (enable)
	msg_data.step_mode = 0; // half steps // 1/16 steps  // default: half steps
	msg_data.dc100_st_en = 1; // disable of motion detection on 100% duty cycle
	msg_data.pwmj_en = 1; // enable pwm jitter to avoid resonance harmonics

        if ( ioctl(file, I2C_RDWR, &wr) <0 )
                return -1;
	close(file);

	return 0;
}

int stepper_reset_position()
{
        int file;
        unsigned char msg_data[1]= {0};
        struct i2c_msg msg = {STEPPER_ADDRESS, 0, 1, (unsigned char*)&msg_data};
        struct i2c_rdwr_ioctl_data wr = { &msg, 1 };
        short reg = 0x86; // ResetPosition
        /* open handle */
        if ((file = open(TWI_DEVICE,O_RDWR)) < 0)
                return -ENODEV;

	msg_data[0] = (unsigned char) reg;

        if ( ioctl(file, I2C_RDWR, &wr) <0 )
                return -1;
	close(file);

	return 0;
}

#define STEPPER_MS_OUTER_MASK 0x4
#define STEPPER_MS_ACC 0x1
#define STEPPER_MS_DEC 0x2
#define STEPPER_MS_MAX 0x3

int busyWaitUntilStopped()
{
  stepper_full_status1_t status1;
    do {
      int res = stepper_read_status1( &status1 );
      if ( res < 0 ) {
	printf( "Error reading status: %i", res);
	return res;
      }
    } while ( status1.motion_status != 0x00 ); // wait until motor is stopped
}

int main (int argc, char **argv)
{
  int res;

  if ( argc == 3 ) { // test
    res = stepper_set_stall_params( 0x0F /* C=800 mA*/,  0x00 /* hold E=673 mA*/);
    stepper_full_status1_t status1;
    char input[80];
    int pos = 0;
    
    if ( res < 0 ) {
      printf( "Error setting stall parameters: %i", res);
      return res;
    }
    
    // reset so we are sure the next command moves complete to the end
    stepper_reset_position();

    // ensure at innermost position
    res = stepper_set_position( -1000 );
    if ( res < 0 ) {
      printf( "Error setting position: %i", res);
      return res;
    }

    busyWaitUntilStopped();

    //    return -1;

    stepper_reset_position();


    do {
      stepper_full_status1_t status1;
      printf( "\nPos: %i. New position (relative), 0 to end: ", pos );
      gets( input );
      
      pos += atoi( input );
      res = stepper_set_position( pos );

      if ( res < 0 ) {
	printf( "Error setting position: %i", res);
	return res;
      }

      printf ( "Waiting for movement...." );
      busyWaitUntilStopped();
      printf( "Movement completed!" );

      /*      res = stepper_read_status1( &status1 );
      if ( res < 0 ) {
	printf( "Error reading status 2: %i", res);
	return res;
      }

      if ( status1.stall_detected ) {
	printf( "\n\n Empty!\n" );

	break;
	}*/

      if (pos > 350 ) {
	printf( "\n\n Empty!\n" );

	break;
      }
    } while (atoi(input) != 0 );
  }

  if ( argc == 2 ) {
    stepper_full_status2_t status2;
    //    res = stepper_set_motor_params( 0x0C /* C=476 mA*/,  0x00 /* hold */);
    //res = stepper_set_motor_params( 0x0F /* C=800 mA*/,  0x00 /* hold E=673 mA*/);
    
    res = stepper_set_stall_params( 0x0F /* C=800 mA*/,  0x00 /* hold E=673 mA*/);

    if ( res < 0 ) {
      printf( "Error setting stall parameters: %i", res);
      return res;
    }

    res = stepper_set_position( atoi(argv[1]) );
    if ( res < 0 ) {
      printf( "Error setting position: %i", res);
      return res;
    }

    res = stepper_read_status2( &status2 );
    if ( res < 0 ) {
      printf( "Error reading status 2: %i", res);
      return res;
    }
    
    printf( "Act_pos: %i, Target_pos: %i, secure_pos: %i, asb_stall: %i, delta_stall_low: %i, delta_stall_high: %i\n",
	    status2.act_pos >> 3, status2.target_pos >> 3, status2.secure_pos_low | (status2.secure_pos_high << 8),
	    status2.abs_stall, status2.delta_stall_low, status2.delta_stall_high );
  }

    // read status
    char buf[7];
    int i;
    stepper_full_status1_t status1;

    res = stepper_read_status1( &status1 );
    if ( res < 0 ) {
      printf( "Error reading status: %i", res);
      return res;
    }
    
    printf( "Stall: %i, step loss: %i, electrical defect: %i, under voltage: %i, temp_info: %i, over_current1: %i, over_current2: %i, charge pump err: %i \n",
	    status1.stall_detected, status1.step_loss, status1.electrical_def, status1.uv2, status1.temp_info, status1.over_current_1, status1.over_current_2, status1.charge_pump_err );

    if ( 0x00 == status1.motion_status ) {
      printf( "Motor stopped\n" );
    }
    else {
      if ( status1.motion_status & STEPPER_MS_OUTER_MASK ) printf( "Outer motion " );
      else printf( "Inner motion " );

      switch (status1.motion_status & 0x3) {
      case STEPPER_MS_ACC:
	printf( "acceleration." );
	break;

      case STEPPER_MS_DEC:
	printf( "deceleration." );
	break;

      case STEPPER_MS_MAX:
	printf( "max. speed." );
	break;
      }
    }

    printf( "\n" );

} // main
